/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2010 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/

#define put_byte(state, c) {state->out_buf[state->out_len++] = (c);}

#define PUT_SHORT(state, w) { \
    put_byte(state, (UBYTE)((w) & 0xff)); \
    put_byte(state, (UBYTE)((unsigned short)(w) >> 8)); \
}
#define SWAP(TYPE, SRC1, SRC2){TYPE TMP_SWAP = (SRC1); (SRC1) = (SRC2); (SRC2) = TMP_SWAP;}

typedef struct block_est {
    short type;
    unsigned int size;
}block_est;
typedef struct freq_table {
    unsigned short frq;
    unsigned short idx;
} freq_table;
typedef struct code_table {
    unsigned short code;
    unsigned short len;
} code_table;
typedef struct fix_huff_tab_ext {
    unsigned int base;
    unsigned short  code;
    unsigned short  len;
} fix_huff_tab_ext;
typedef struct ct_data {
    union {
        unsigned short  freq;
        unsigned short  code;
    } fc;
    union {
        unsigned short  dad;
        unsigned short  len;
    } dl;
}ct_data;

static const ct_data static_len_code[288] = {
   {{ 12},{  8}}, {{140},{  8}}, {{ 76},{  8}}, {{204},{  8}}, {{ 44},{  8}},
   {{172},{  8}}, {{108},{  8}}, {{236},{  8}}, {{ 28},{  8}}, {{156},{  8}},
   {{ 92},{  8}}, {{220},{  8}}, {{ 60},{  8}}, {{188},{  8}}, {{124},{  8}},
   {{252},{  8}}, {{  2},{  8}}, {{130},{  8}}, {{ 66},{  8}}, {{194},{  8}},
   {{ 34},{  8}}, {{162},{  8}}, {{ 98},{  8}}, {{226},{  8}}, {{ 18},{  8}},
   {{146},{  8}}, {{ 82},{  8}}, {{210},{  8}}, {{ 50},{  8}}, {{178},{  8}},
   {{114},{  8}}, {{242},{  8}}, {{ 10},{  8}}, {{138},{  8}}, {{ 74},{  8}},
   {{202},{  8}}, {{ 42},{  8}}, {{170},{  8}}, {{106},{  8}}, {{234},{  8}},
   {{ 26},{  8}}, {{154},{  8}}, {{ 90},{  8}}, {{218},{  8}}, {{ 58},{  8}},
   {{186},{  8}}, {{122},{  8}}, {{250},{  8}}, {{  6},{  8}}, {{134},{  8}},
   {{ 70},{  8}}, {{198},{  8}}, {{ 38},{  8}}, {{166},{  8}}, {{102},{  8}},
   {{230},{  8}}, {{ 22},{  8}}, {{150},{  8}}, {{ 86},{  8}}, {{214},{  8}},
   {{ 54},{  8}}, {{182},{  8}}, {{118},{  8}}, {{246},{  8}}, {{ 14},{  8}},
   {{142},{  8}}, {{ 78},{  8}}, {{206},{  8}}, {{ 46},{  8}}, {{174},{  8}},
   {{110},{  8}}, {{238},{  8}}, {{ 30},{  8}}, {{158},{  8}}, {{ 94},{  8}},
   {{222},{  8}}, {{ 62},{  8}}, {{190},{  8}}, {{126},{  8}}, {{254},{  8}},
   {{  1},{  8}}, {{129},{  8}}, {{ 65},{  8}}, {{193},{  8}}, {{ 33},{  8}},
   {{161},{  8}}, {{ 97},{  8}}, {{225},{  8}}, {{ 17},{  8}}, {{145},{  8}},
   {{ 81},{  8}}, {{209},{  8}}, {{ 49},{  8}}, {{177},{  8}}, {{113},{  8}},
   {{241},{  8}}, {{  9},{  8}}, {{137},{  8}}, {{ 73},{  8}}, {{201},{  8}},
   {{ 41},{  8}}, {{169},{  8}}, {{105},{  8}}, {{233},{  8}}, {{ 25},{  8}},
   {{153},{  8}}, {{ 89},{  8}}, {{217},{  8}}, {{ 57},{  8}}, {{185},{  8}},
   {{121},{  8}}, {{249},{  8}}, {{  5},{  8}}, {{133},{  8}}, {{ 69},{  8}},
   {{197},{  8}}, {{ 37},{  8}}, {{165},{  8}}, {{101},{  8}}, {{229},{  8}},
   {{ 21},{  8}}, {{149},{  8}}, {{ 85},{  8}}, {{213},{  8}}, {{ 53},{  8}},
   {{181},{  8}}, {{117},{  8}}, {{245},{  8}}, {{ 13},{  8}}, {{141},{  8}},
   {{ 77},{  8}}, {{205},{  8}}, {{ 45},{  8}}, {{173},{  8}}, {{109},{  8}},
   {{237},{  8}}, {{ 29},{  8}}, {{157},{  8}}, {{ 93},{  8}}, {{221},{  8}},
   {{ 61},{  8}}, {{189},{  8}}, {{125},{  8}}, {{253},{  8}}, {{ 19},{  9}},
   {{275},{  9}}, {{147},{  9}}, {{403},{  9}}, {{ 83},{  9}}, {{339},{  9}},
   {{211},{  9}}, {{467},{  9}}, {{ 51},{  9}}, {{307},{  9}}, {{179},{  9}},
   {{435},{  9}}, {{115},{  9}}, {{371},{  9}}, {{243},{  9}}, {{499},{  9}},
   {{ 11},{  9}}, {{267},{  9}}, {{139},{  9}}, {{395},{  9}}, {{ 75},{  9}},
   {{331},{  9}}, {{203},{  9}}, {{459},{  9}}, {{ 43},{  9}}, {{299},{  9}},
   {{171},{  9}}, {{427},{  9}}, {{107},{  9}}, {{363},{  9}}, {{235},{  9}},
   {{491},{  9}}, {{ 27},{  9}}, {{283},{  9}}, {{155},{  9}}, {{411},{  9}},
   {{ 91},{  9}}, {{347},{  9}}, {{219},{  9}}, {{475},{  9}}, {{ 59},{  9}},
   {{315},{  9}}, {{187},{  9}}, {{443},{  9}}, {{123},{  9}}, {{379},{  9}},
   {{251},{  9}}, {{507},{  9}}, {{  7},{  9}}, {{263},{  9}}, {{135},{  9}},
   {{391},{  9}}, {{ 71},{  9}}, {{327},{  9}}, {{199},{  9}}, {{455},{  9}},
   {{ 39},{  9}}, {{295},{  9}}, {{167},{  9}}, {{423},{  9}}, {{103},{  9}},
   {{359},{  9}}, {{231},{  9}}, {{487},{  9}}, {{ 23},{  9}}, {{279},{  9}},
   {{151},{  9}}, {{407},{  9}}, {{ 87},{  9}}, {{343},{  9}}, {{215},{  9}},
   {{471},{  9}}, {{ 55},{  9}}, {{311},{  9}}, {{183},{  9}}, {{439},{  9}},
   {{119},{  9}}, {{375},{  9}}, {{247},{  9}}, {{503},{  9}}, {{ 15},{  9}},
   {{271},{  9}}, {{143},{  9}}, {{399},{  9}}, {{ 79},{  9}}, {{335},{  9}},
   {{207},{  9}}, {{463},{  9}}, {{ 47},{  9}}, {{303},{  9}}, {{175},{  9}},
   {{431},{  9}}, {{111},{  9}}, {{367},{  9}}, {{239},{  9}}, {{495},{  9}},
   {{ 31},{  9}}, {{287},{  9}}, {{159},{  9}}, {{415},{  9}}, {{ 95},{  9}},
   {{351},{  9}}, {{223},{  9}}, {{479},{  9}}, {{ 63},{  9}}, {{319},{  9}},
   {{191},{  9}}, {{447},{  9}}, {{127},{  9}}, {{383},{  9}}, {{255},{  9}},
   {{511},{  9}}, {{  0},{  7}}, {{ 64},{  7}}, {{ 32},{  7}}, {{ 96},{  7}},
   {{ 16},{  7}}, {{ 80},{  7}}, {{ 48},{  7}}, {{112},{  7}}, {{  8},{  7}},
   {{ 72},{  7}}, {{ 40},{  7}}, {{104},{  7}}, {{ 24},{  7}}, {{ 88},{  7}},
   {{ 56},{  7}}, {{120},{  7}}, {{  4},{  7}}, {{ 68},{  7}}, {{ 36},{  7}},
   {{100},{  7}}, {{ 20},{  7}}, {{ 84},{  7}}, {{ 52},{  7}}, {{116},{  7}},
   {{  3},{  8}}, {{131},{  8}}, {{ 67},{  8}}, {{195},{  8}}, {{ 35},{  8}},
   {{163},{  8}}, {{ 99},{  8}}, {{227},{  8}}
};

static const ct_data static_oft_code[30] = {
   {{ 0},{ 5}}, {{16},{ 5}}, {{ 8},{ 5}}, {{24},{ 5}}, {{ 4},{ 5}},
   {{20},{ 5}}, {{12},{ 5}}, {{28},{ 5}}, {{ 2},{ 5}}, {{18},{ 5}},
   {{10},{ 5}}, {{26},{ 5}}, {{ 6},{ 5}}, {{22},{ 5}}, {{14},{ 5}},
   {{30},{ 5}}, {{ 1},{ 5}}, {{17},{ 5}}, {{ 9},{ 5}}, {{25},{ 5}},
   {{ 5},{ 5}}, {{21},{ 5}}, {{13},{ 5}}, {{29},{ 5}}, {{ 3},{ 5}},
   {{19},{ 5}}, {{11},{ 5}}, {{27},{ 5}}, {{ 7},{ 5}}, {{23},{ 5}}
};

static const code_table len_excode_table[256] = {
  {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0},
  {  0, 1}, {  1, 1}, {  0, 1}, {  1, 1}, {  0, 1}, {  1, 1}, {  0, 1}, {  1, 1},
  {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2}, {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2},
  {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2}, {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, { 31, 5},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, { 31, 5},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, { 31, 5},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, {  0, 0}
};

static const code_table oft_excode_table[256] = {
  {  0, 0}, {  0, 0}, {  0, 0}, {  0, 0}, {  0, 1}, {  1, 1}, {  0, 1}, {  1, 1},
  {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2}, {  0, 2}, {  1, 2}, {  2, 2}, {  3, 2},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 3}, {  1, 3}, {  2, 3}, {  3, 3}, {  4, 3}, {  5, 3}, {  6, 3}, {  7, 3},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 4}, {  1, 4}, {  2, 4}, {  3, 4}, {  4, 4}, {  5, 4}, {  6, 4}, {  7, 4},
  {  8, 4}, {  9, 4}, { 10, 4}, { 11, 4}, { 12, 4}, { 13, 4}, { 14, 4}, { 15, 4},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, { 31, 5},
  {  0, 5}, {  1, 5}, {  2, 5}, {  3, 5}, {  4, 5}, {  5, 5}, {  6, 5}, {  7, 5},
  {  8, 5}, {  9, 5}, { 10, 5}, { 11, 5}, { 12, 5}, { 13, 5}, { 14, 5}, { 15, 5},
  { 16, 5}, { 17, 5}, { 18, 5}, { 19, 5}, { 20, 5}, { 21, 5}, { 22, 5}, { 23, 5},
  { 24, 5}, { 25, 5}, { 26, 5}, { 27, 5}, { 28, 5}, { 29, 5}, { 30, 5}, { 31, 5},
  {  0, 6}, {  1, 6}, {  2, 6}, {  3, 6}, {  4, 6}, {  5, 6}, {  6, 6}, {  7, 6},
  {  8, 6}, {  9, 6}, { 10, 6}, { 11, 6}, { 12, 6}, { 13, 6}, { 14, 6}, { 15, 6},
  { 16, 6}, { 17, 6}, { 18, 6}, { 19, 6}, { 20, 6}, { 21, 6}, { 22, 6}, { 23, 6},
  { 24, 6}, { 25, 6}, { 26, 6}, { 27, 6}, { 28, 6}, { 29, 6}, { 30, 6}, { 31, 6},
  { 32, 6}, { 33, 6}, { 34, 6}, { 35, 6}, { 36, 6}, { 37, 6}, { 38, 6}, { 39, 6},
  { 40, 6}, { 41, 6}, { 42, 6}, { 43, 6}, { 44, 6}, { 45, 6}, { 46, 6}, { 47, 6},
  { 48, 6}, { 49, 6}, { 50, 6}, { 51, 6}, { 52, 6}, { 53, 6}, { 54, 6}, { 55, 6},
  { 56, 6}, { 57, 6}, { 58, 6}, { 59, 6}, { 60, 6}, { 61, 6}, { 62, 6}, { 63, 6},
  {  0, 6}, {  1, 6}, {  2, 6}, {  3, 6}, {  4, 6}, {  5, 6}, {  6, 6}, {  7, 6},
  {  8, 6}, {  9, 6}, { 10, 6}, { 11, 6}, { 12, 6}, { 13, 6}, { 14, 6}, { 15, 6},
  { 16, 6}, { 17, 6}, { 18, 6}, { 19, 6}, { 20, 6}, { 21, 6}, { 22, 6}, { 23, 6},
  { 24, 6}, { 25, 6}, { 26, 6}, { 27, 6}, { 28, 6}, { 29, 6}, { 30, 6}, { 31, 6},
  { 32, 6}, { 33, 6}, { 34, 6}, { 35, 6}, { 36, 6}, { 37, 6}, { 38, 6}, { 39, 6},
  { 40, 6}, { 41, 6}, { 42, 6}, { 43, 6}, { 44, 6}, { 45, 6}, { 46, 6}, { 47, 6},
  { 48, 6}, { 49, 6}, { 50, 6}, { 51, 6}, { 52, 6}, { 53, 6}, { 54, 6}, { 55, 6},
  { 56, 6}, { 57, 6}, { 58, 6}, { 59, 6}, { 60, 6}, { 61, 6}, { 62, 6}, { 63, 6}
};

static const code_table l_table[256] = { /* length table */
  { 0x0040,  7 }, { 0x0020,  7 }, { 0x0060,  7 }, { 0x0010,  7 },
  { 0x0050,  7 }, { 0x0030,  7 }, { 0x0070,  7 }, { 0x0008,  7 },
  { 0x0048,  8 }, { 0x00c8,  8 }, { 0x0028,  8 }, { 0x00a8,  8 },
  { 0x0068,  8 }, { 0x00e8,  8 }, { 0x0018,  8 }, { 0x0098,  8 },
  { 0x0058,  9 }, { 0x00d8,  9 }, { 0x0158,  9 }, { 0x01d8,  9 },
  { 0x0038,  9 }, { 0x00b8,  9 }, { 0x0138,  9 }, { 0x01b8,  9 },
  { 0x0078,  9 }, { 0x00f8,  9 }, { 0x0178,  9 }, { 0x01f8,  9 },
  { 0x0004,  9 }, { 0x0084,  9 }, { 0x0104,  9 }, { 0x0184,  9 },
  { 0x0044, 10 }, { 0x00c4, 10 }, { 0x0144, 10 }, { 0x01c4, 10 },
  { 0x0244, 10 }, { 0x02c4, 10 }, { 0x0344, 10 }, { 0x03c4, 10 },
  { 0x0024, 10 }, { 0x00a4, 10 }, { 0x0124, 10 }, { 0x01a4, 10 },
  { 0x0224, 10 }, { 0x02a4, 10 }, { 0x0324, 10 }, { 0x03a4, 10 },
  { 0x0064, 10 }, { 0x00e4, 10 }, { 0x0164, 10 }, { 0x01e4, 10 },
  { 0x0264, 10 }, { 0x02e4, 10 }, { 0x0364, 10 }, { 0x03e4, 10 },
  { 0x0014, 10 }, { 0x0094, 10 }, { 0x0114, 10 }, { 0x0194, 10 },
  { 0x0214, 10 }, { 0x0294, 10 }, { 0x0314, 10 }, { 0x0394, 10 },
  { 0x0054, 11 }, { 0x00d4, 11 }, { 0x0154, 11 }, { 0x01d4, 11 },
  { 0x0254, 11 }, { 0x02d4, 11 }, { 0x0354, 11 }, { 0x03d4, 11 },
  { 0x0454, 11 }, { 0x04d4, 11 }, { 0x0554, 11 }, { 0x05d4, 11 },
  { 0x0654, 11 }, { 0x06d4, 11 }, { 0x0754, 11 }, { 0x07d4, 11 },
  { 0x0034, 11 }, { 0x00b4, 11 }, { 0x0134, 11 }, { 0x01b4, 11 },
  { 0x0234, 11 }, { 0x02b4, 11 }, { 0x0334, 11 }, { 0x03b4, 11 },
  { 0x0434, 11 }, { 0x04b4, 11 }, { 0x0534, 11 }, { 0x05b4, 11 },
  { 0x0634, 11 }, { 0x06b4, 11 }, { 0x0734, 11 }, { 0x07b4, 11 },
  { 0x0074, 11 }, { 0x00f4, 11 }, { 0x0174, 11 }, { 0x01f4, 11 },
  { 0x0274, 11 }, { 0x02f4, 11 }, { 0x0374, 11 }, { 0x03f4, 11 },
  { 0x0474, 11 }, { 0x04f4, 11 }, { 0x0574, 11 }, { 0x05f4, 11 },
  { 0x0674, 11 }, { 0x06f4, 11 }, { 0x0774, 11 }, { 0x07f4, 11 },
  { 0x0003, 12 }, { 0x0103, 12 }, { 0x0203, 12 }, { 0x0303, 12 },
  { 0x0403, 12 }, { 0x0503, 12 }, { 0x0603, 12 }, { 0x0703, 12 },
  { 0x0803, 12 }, { 0x0903, 12 }, { 0x0a03, 12 }, { 0x0b03, 12 },
  { 0x0c03, 12 }, { 0x0d03, 12 }, { 0x0e03, 12 }, { 0x0f03, 12 },
  { 0x0083, 13 }, { 0x0183, 13 }, { 0x0283, 13 }, { 0x0383, 13 },
  { 0x0483, 13 }, { 0x0583, 13 }, { 0x0683, 13 }, { 0x0783, 13 },
  { 0x0883, 13 }, { 0x0983, 13 }, { 0x0a83, 13 }, { 0x0b83, 13 },
  { 0x0c83, 13 }, { 0x0d83, 13 }, { 0x0e83, 13 }, { 0x0f83, 13 },
  { 0x1083, 13 }, { 0x1183, 13 }, { 0x1283, 13 }, { 0x1383, 13 },
  { 0x1483, 13 }, { 0x1583, 13 }, { 0x1683, 13 }, { 0x1783, 13 },
  { 0x1883, 13 }, { 0x1983, 13 }, { 0x1a83, 13 }, { 0x1b83, 13 },
  { 0x1c83, 13 }, { 0x1d83, 13 }, { 0x1e83, 13 }, { 0x1f83, 13 },
  { 0x0043, 13 }, { 0x0143, 13 }, { 0x0243, 13 }, { 0x0343, 13 },
  { 0x0443, 13 }, { 0x0543, 13 }, { 0x0643, 13 }, { 0x0743, 13 },
  { 0x0843, 13 }, { 0x0943, 13 }, { 0x0a43, 13 }, { 0x0b43, 13 },
  { 0x0c43, 13 }, { 0x0d43, 13 }, { 0x0e43, 13 }, { 0x0f43, 13 },
  { 0x1043, 13 }, { 0x1143, 13 }, { 0x1243, 13 }, { 0x1343, 13 },
  { 0x1443, 13 }, { 0x1543, 13 }, { 0x1643, 13 }, { 0x1743, 13 },
  { 0x1843, 13 }, { 0x1943, 13 }, { 0x1a43, 13 }, { 0x1b43, 13 },
  { 0x1c43, 13 }, { 0x1d43, 13 }, { 0x1e43, 13 }, { 0x1f43, 13 },
  { 0x00c3, 13 }, { 0x01c3, 13 }, { 0x02c3, 13 }, { 0x03c3, 13 },
  { 0x04c3, 13 }, { 0x05c3, 13 }, { 0x06c3, 13 }, { 0x07c3, 13 },
  { 0x08c3, 13 }, { 0x09c3, 13 }, { 0x0ac3, 13 }, { 0x0bc3, 13 },
  { 0x0cc3, 13 }, { 0x0dc3, 13 }, { 0x0ec3, 13 }, { 0x0fc3, 13 },
  { 0x10c3, 13 }, { 0x11c3, 13 }, { 0x12c3, 13 }, { 0x13c3, 13 },
  { 0x14c3, 13 }, { 0x15c3, 13 }, { 0x16c3, 13 }, { 0x17c3, 13 },
  { 0x18c3, 13 }, { 0x19c3, 13 }, { 0x1ac3, 13 }, { 0x1bc3, 13 },
  { 0x1cc3, 13 }, { 0x1dc3, 13 }, { 0x1ec3, 13 }, { 0x1fc3, 13 },
  { 0x0023, 13 }, { 0x0123, 13 }, { 0x0223, 13 }, { 0x0323, 13 },
  { 0x0423, 13 }, { 0x0523, 13 }, { 0x0623, 13 }, { 0x0723, 13 },
  { 0x0823, 13 }, { 0x0923, 13 }, { 0x0a23, 13 }, { 0x0b23, 13 },
  { 0x0c23, 13 }, { 0x0d23, 13 }, { 0x0e23, 13 }, { 0x0f23, 13 },
  { 0x1023, 13 }, { 0x1123, 13 }, { 0x1223, 13 }, { 0x1323, 13 },
  { 0x1423, 13 }, { 0x1523, 13 }, { 0x1623, 13 }, { 0x1723, 13 },
  { 0x1823, 13 }, { 0x1923, 13 }, { 0x1a23, 13 }, { 0x1b23, 13 },
  { 0x1c23, 13 }, { 0x1d23, 13 }, { 0x1e23, 13 }, { 0x00a3,  8 }
};


static const code_table c_table[256] = {  /* literal table */
  { 0x000c, 8 }, { 0x008c, 8 }, { 0x004c, 8 }, { 0x00cc, 8 },
  { 0x002c, 8 }, { 0x00ac, 8 }, { 0x006c, 8 }, { 0x00ec, 8 },
  { 0x001c, 8 }, { 0x009c, 8 }, { 0x005c, 8 }, { 0x00dc, 8 },
  { 0x003c, 8 }, { 0x00bc, 8 }, { 0x007c, 8 }, { 0x00fc, 8 },
  { 0x0002, 8 }, { 0x0082, 8 }, { 0x0042, 8 }, { 0x00c2, 8 },
  { 0x0022, 8 }, { 0x00a2, 8 }, { 0x0062, 8 }, { 0x00e2, 8 },
  { 0x0012, 8 }, { 0x0092, 8 }, { 0x0052, 8 }, { 0x00d2, 8 },
  { 0x0032, 8 }, { 0x00b2, 8 }, { 0x0072, 8 }, { 0x00f2, 8 },
  { 0x000a, 8 }, { 0x008a, 8 }, { 0x004a, 8 }, { 0x00ca, 8 },
  { 0x002a, 8 }, { 0x00aa, 8 }, { 0x006a, 8 }, { 0x00ea, 8 },
  { 0x001a, 8 }, { 0x009a, 8 }, { 0x005a, 8 }, { 0x00da, 8 },
  { 0x003a, 8 }, { 0x00ba, 8 }, { 0x007a, 8 }, { 0x00fa, 8 },
  { 0x0006, 8 }, { 0x0086, 8 }, { 0x0046, 8 }, { 0x00c6, 8 },
  { 0x0026, 8 }, { 0x00a6, 8 }, { 0x0066, 8 }, { 0x00e6, 8 },
  { 0x0016, 8 }, { 0x0096, 8 }, { 0x0056, 8 }, { 0x00d6, 8 },
  { 0x0036, 8 }, { 0x00b6, 8 }, { 0x0076, 8 }, { 0x00f6, 8 },
  { 0x000e, 8 }, { 0x008e, 8 }, { 0x004e, 8 }, { 0x00ce, 8 },
  { 0x002e, 8 }, { 0x00ae, 8 }, { 0x006e, 8 }, { 0x00ee, 8 },
  { 0x001e, 8 }, { 0x009e, 8 }, { 0x005e, 8 }, { 0x00de, 8 },
  { 0x003e, 8 }, { 0x00be, 8 }, { 0x007e, 8 }, { 0x00fe, 8 },
  { 0x0001, 8 }, { 0x0081, 8 }, { 0x0041, 8 }, { 0x00c1, 8 },
  { 0x0021, 8 }, { 0x00a1, 8 }, { 0x0061, 8 }, { 0x00e1, 8 },
  { 0x0011, 8 }, { 0x0091, 8 }, { 0x0051, 8 }, { 0x00d1, 8 },
  { 0x0031, 8 }, { 0x00b1, 8 }, { 0x0071, 8 }, { 0x00f1, 8 },
  { 0x0009, 8 }, { 0x0089, 8 }, { 0x0049, 8 }, { 0x00c9, 8 },
  { 0x0029, 8 }, { 0x00a9, 8 }, { 0x0069, 8 }, { 0x00e9, 8 },
  { 0x0019, 8 }, { 0x0099, 8 }, { 0x0059, 8 }, { 0x00d9, 8 },
  { 0x0039, 8 }, { 0x00b9, 8 }, { 0x0079, 8 }, { 0x00f9, 8 },
  { 0x0005, 8 }, { 0x0085, 8 }, { 0x0045, 8 }, { 0x00c5, 8 },
  { 0x0025, 8 }, { 0x00a5, 8 }, { 0x0065, 8 }, { 0x00e5, 8 },
  { 0x0015, 8 }, { 0x0095, 8 }, { 0x0055, 8 }, { 0x00d5, 8 },
  { 0x0035, 8 }, { 0x00b5, 8 }, { 0x0075, 8 }, { 0x00f5, 8 },
  { 0x000d, 8 }, { 0x008d, 8 }, { 0x004d, 8 }, { 0x00cd, 8 },
  { 0x002d, 8 }, { 0x00ad, 8 }, { 0x006d, 8 }, { 0x00ed, 8 },
  { 0x001d, 8 }, { 0x009d, 8 }, { 0x005d, 8 }, { 0x00dd, 8 },
  { 0x003d, 8 }, { 0x00bd, 8 }, { 0x007d, 8 }, { 0x00fd, 8 },
  { 0x0013, 9 }, { 0x0113, 9 }, { 0x0093, 9 }, { 0x0193, 9 },
  { 0x0053, 9 }, { 0x0153, 9 }, { 0x00d3, 9 }, { 0x01d3, 9 },
  { 0x0033, 9 }, { 0x0133, 9 }, { 0x00b3, 9 }, { 0x01b3, 9 },
  { 0x0073, 9 }, { 0x0173, 9 }, { 0x00f3, 9 }, { 0x01f3, 9 },
  { 0x000b, 9 }, { 0x010b, 9 }, { 0x008b, 9 }, { 0x018b, 9 },
  { 0x004b, 9 }, { 0x014b, 9 }, { 0x00cb, 9 }, { 0x01cb, 9 },
  { 0x002b, 9 }, { 0x012b, 9 }, { 0x00ab, 9 }, { 0x01ab, 9 },
  { 0x006b, 9 }, { 0x016b, 9 }, { 0x00eb, 9 }, { 0x01eb, 9 },
  { 0x001b, 9 }, { 0x011b, 9 }, { 0x009b, 9 }, { 0x019b, 9 },
  { 0x005b, 9 }, { 0x015b, 9 }, { 0x00db, 9 }, { 0x01db, 9 },
  { 0x003b, 9 }, { 0x013b, 9 }, { 0x00bb, 9 }, { 0x01bb, 9 },
  { 0x007b, 9 }, { 0x017b, 9 }, { 0x00fb, 9 }, { 0x01fb, 9 },
  { 0x0007, 9 }, { 0x0107, 9 }, { 0x0087, 9 }, { 0x0187, 9 },
  { 0x0047, 9 }, { 0x0147, 9 }, { 0x00c7, 9 }, { 0x01c7, 9 },
  { 0x0027, 9 }, { 0x0127, 9 }, { 0x00a7, 9 }, { 0x01a7, 9 },
  { 0x0067, 9 }, { 0x0167, 9 }, { 0x00e7, 9 }, { 0x01e7, 9 },
  { 0x0017, 9 }, { 0x0117, 9 }, { 0x0097, 9 }, { 0x0197, 9 },
  { 0x0057, 9 }, { 0x0157, 9 }, { 0x00d7, 9 }, { 0x01d7, 9 },
  { 0x0037, 9 }, { 0x0137, 9 }, { 0x00b7, 9 }, { 0x01b7, 9 },
  { 0x0077, 9 }, { 0x0177, 9 }, { 0x00f7, 9 }, { 0x01f7, 9 },
  { 0x000f, 9 }, { 0x010f, 9 }, { 0x008f, 9 }, { 0x018f, 9 },
  { 0x004f, 9 }, { 0x014f, 9 }, { 0x00cf, 9 }, { 0x01cf, 9 },
  { 0x002f, 9 }, { 0x012f, 9 }, { 0x00af, 9 }, { 0x01af, 9 },
  { 0x006f, 9 }, { 0x016f, 9 }, { 0x00ef, 9 }, { 0x01ef, 9 },
  { 0x001f, 9 }, { 0x011f, 9 }, { 0x009f, 9 }, { 0x019f, 9 },
  { 0x005f, 9 }, { 0x015f, 9 }, { 0x00df, 9 }, { 0x01df, 9 },
  { 0x003f, 9 }, { 0x013f, 9 }, { 0x00bf, 9 }, { 0x01bf, 9 },
  { 0x007f, 9 }, { 0x017f, 9 }, { 0x00ff, 9 }, { 0x01ff, 9 }
};

static const code_table o_table_l[256] = { /* offset table */
  { 0x0000,  5 }, { 0x0010,  5 }, { 0x0008,  5 }, { 0x0018,  5 },
  { 0x0004,  6 }, { 0x0024,  6 }, { 0x0014,  6 }, { 0x0034,  6 },
  { 0x000c,  7 }, { 0x002c,  7 }, { 0x004c,  7 }, { 0x006c,  7 },
  { 0x001c,  7 }, { 0x003c,  7 }, { 0x005c,  7 }, { 0x007c,  7 },
  { 0x0002,  8 }, { 0x0022,  8 }, { 0x0042,  8 }, { 0x0062,  8 },
  { 0x0082,  8 }, { 0x00a2,  8 }, { 0x00c2,  8 }, { 0x00e2,  8 },
  { 0x0012,  8 }, { 0x0032,  8 }, { 0x0052,  8 }, { 0x0072,  8 },
  { 0x0092,  8 }, { 0x00b2,  8 }, { 0x00d2,  8 }, { 0x00f2,  8 },
  { 0x000a,  9 }, { 0x002a,  9 }, { 0x004a,  9 }, { 0x006a,  9 },
  { 0x008a,  9 }, { 0x00aa,  9 }, { 0x00ca,  9 }, { 0x00ea,  9 },
  { 0x010a,  9 }, { 0x012a,  9 }, { 0x014a,  9 }, { 0x016a,  9 },
  { 0x018a,  9 }, { 0x01aa,  9 }, { 0x01ca,  9 }, { 0x01ea,  9 },
  { 0x001a,  9 }, { 0x003a,  9 }, { 0x005a,  9 }, { 0x007a,  9 },
  { 0x009a,  9 }, { 0x00ba,  9 }, { 0x00da,  9 }, { 0x00fa,  9 },
  { 0x011a,  9 }, { 0x013a,  9 }, { 0x015a,  9 }, { 0x017a,  9 },
  { 0x019a,  9 }, { 0x01ba,  9 }, { 0x01da,  9 }, { 0x01fa,  9 },
  { 0x0006, 10 }, { 0x0026, 10 }, { 0x0046, 10 }, { 0x0066, 10 },
  { 0x0086, 10 }, { 0x00a6, 10 }, { 0x00c6, 10 }, { 0x00e6, 10 },
  { 0x0106, 10 }, { 0x0126, 10 }, { 0x0146, 10 }, { 0x0166, 10 },
  { 0x0186, 10 }, { 0x01a6, 10 }, { 0x01c6, 10 }, { 0x01e6, 10 },
  { 0x0206, 10 }, { 0x0226, 10 }, { 0x0246, 10 }, { 0x0266, 10 },
  { 0x0286, 10 }, { 0x02a6, 10 }, { 0x02c6, 10 }, { 0x02e6, 10 },
  { 0x0306, 10 }, { 0x0326, 10 }, { 0x0346, 10 }, { 0x0366, 10 },
  { 0x0386, 10 }, { 0x03a6, 10 }, { 0x03c6, 10 }, { 0x03e6, 10 },
  { 0x0016, 10 }, { 0x0036, 10 }, { 0x0056, 10 }, { 0x0076, 10 },
  { 0x0096, 10 }, { 0x00b6, 10 }, { 0x00d6, 10 }, { 0x00f6, 10 },
  { 0x0116, 10 }, { 0x0136, 10 }, { 0x0156, 10 }, { 0x0176, 10 },
  { 0x0196, 10 }, { 0x01b6, 10 }, { 0x01d6, 10 }, { 0x01f6, 10 },
  { 0x0216, 10 }, { 0x0236, 10 }, { 0x0256, 10 }, { 0x0276, 10 },
  { 0x0296, 10 }, { 0x02b6, 10 }, { 0x02d6, 10 }, { 0x02f6, 10 },
  { 0x0316, 10 }, { 0x0336, 10 }, { 0x0356, 10 }, { 0x0376, 10 },
  { 0x0396, 10 }, { 0x03b6, 10 }, { 0x03d6, 10 }, { 0x03f6, 10 },
  { 0x000e, 11 }, { 0x002e, 11 }, { 0x004e, 11 }, { 0x006e, 11 },
  { 0x008e, 11 }, { 0x00ae, 11 }, { 0x00ce, 11 }, { 0x00ee, 11 },
  { 0x010e, 11 }, { 0x012e, 11 }, { 0x014e, 11 }, { 0x016e, 11 },
  { 0x018e, 11 }, { 0x01ae, 11 }, { 0x01ce, 11 }, { 0x01ee, 11 },
  { 0x020e, 11 }, { 0x022e, 11 }, { 0x024e, 11 }, { 0x026e, 11 },
  { 0x028e, 11 }, { 0x02ae, 11 }, { 0x02ce, 11 }, { 0x02ee, 11 },
  { 0x030e, 11 }, { 0x032e, 11 }, { 0x034e, 11 }, { 0x036e, 11 },
  { 0x038e, 11 }, { 0x03ae, 11 }, { 0x03ce, 11 }, { 0x03ee, 11 },
  { 0x040e, 11 }, { 0x042e, 11 }, { 0x044e, 11 }, { 0x046e, 11 },
  { 0x048e, 11 }, { 0x04ae, 11 }, { 0x04ce, 11 }, { 0x04ee, 11 },
  { 0x050e, 11 }, { 0x052e, 11 }, { 0x054e, 11 }, { 0x056e, 11 },
  { 0x058e, 11 }, { 0x05ae, 11 }, { 0x05ce, 11 }, { 0x05ee, 11 },
  { 0x060e, 11 }, { 0x062e, 11 }, { 0x064e, 11 }, { 0x066e, 11 },
  { 0x068e, 11 }, { 0x06ae, 11 }, { 0x06ce, 11 }, { 0x06ee, 11 },
  { 0x070e, 11 }, { 0x072e, 11 }, { 0x074e, 11 }, { 0x076e, 11 },
  { 0x078e, 11 }, { 0x07ae, 11 }, { 0x07ce, 11 }, { 0x07ee, 11 },
  { 0x001e, 11 }, { 0x003e, 11 }, { 0x005e, 11 }, { 0x007e, 11 },
  { 0x009e, 11 }, { 0x00be, 11 }, { 0x00de, 11 }, { 0x00fe, 11 },
  { 0x011e, 11 }, { 0x013e, 11 }, { 0x015e, 11 }, { 0x017e, 11 },
  { 0x019e, 11 }, { 0x01be, 11 }, { 0x01de, 11 }, { 0x01fe, 11 },
  { 0x021e, 11 }, { 0x023e, 11 }, { 0x025e, 11 }, { 0x027e, 11 },
  { 0x029e, 11 }, { 0x02be, 11 }, { 0x02de, 11 }, { 0x02fe, 11 },
  { 0x031e, 11 }, { 0x033e, 11 }, { 0x035e, 11 }, { 0x037e, 11 },
  { 0x039e, 11 }, { 0x03be, 11 }, { 0x03de, 11 }, { 0x03fe, 11 },
  { 0x041e, 11 }, { 0x043e, 11 }, { 0x045e, 11 }, { 0x047e, 11 },
  { 0x049e, 11 }, { 0x04be, 11 }, { 0x04de, 11 }, { 0x04fe, 11 },
  { 0x051e, 11 }, { 0x053e, 11 }, { 0x055e, 11 }, { 0x057e, 11 },
  { 0x059e, 11 }, { 0x05be, 11 }, { 0x05de, 11 }, { 0x05fe, 11 },
  { 0x061e, 11 }, { 0x063e, 11 }, { 0x065e, 11 }, { 0x067e, 11 },
  { 0x069e, 11 }, { 0x06be, 11 }, { 0x06de, 11 }, { 0x06fe, 11 },
  { 0x071e, 11 }, { 0x073e, 11 }, { 0x075e, 11 }, { 0x077e, 11 },
  { 0x079e, 11 }, { 0x07be, 11 }, { 0x07de, 11 }, { 0x07fe, 11 }
};

static const fix_huff_tab_ext o_table_h[15] = { /* offset table */
  { 0x0101, 0x01,  7 }, { 0x0181, 0x11,  7 }, { 0x0201, 0x09,  8 }, { 0x0301, 0x19,  8 },
  { 0x0401, 0x05,  9 }, { 0x0601, 0x15,  9 }, { 0x0801, 0x0d, 10 }, { 0x0C01, 0x1d, 10 },
  { 0x1001, 0x03, 11 }, { 0x1801, 0x13, 11 }, { 0x2001, 0x0b, 12 }, { 0x3001, 0x1b, 12 },
  { 0x4001, 0x07, 13 }, { 0x6001, 0x17, 13 }, { 0x8001, 0x0f, 14 }
};

static const UBYTE l_code_tab[256]= {
    0x0,  0x1,  0x2,  0x3,  0x4,  0x5,  0x6,  0x7,  0x8,  0x8,  0x9,  0x9,  0xa,  0xa,  0xb,  0xb,  0xc,  0xc,  0xc,  0xc,
    0xd,  0xd,  0xd,  0xd,  0xe,  0xe,  0xe,  0xe,  0xf,  0xf,  0xf,  0xf, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
   0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x13, 0x13, 0x13, 0x13,
   0x13, 0x13, 0x13, 0x13, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14,
   0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x16, 0x16, 0x16, 0x16,
   0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17,
   0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
   0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
   0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19,
   0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a,
   0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a,
   0x1a, 0x1a, 0x1a, 0x1a, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
   0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1c
};

static const UBYTE d_code_tab[512] = {
    0x0,  0x1,  0x2,  0x3,  0x4,  0x4,  0x5,  0x5,  0x6,  0x6,  0x6,  0x6,  0x7,  0x7,  0x7,  0x7,  0x8,  0x8,  0x8,  0x8,
    0x8,  0x8,  0x8,  0x8,  0x9,  0x9,  0x9,  0x9,  0x9,  0x9,  0x9,  0x9,  0xa,  0xa,  0xa,  0xa,  0xa,  0xa,  0xa,  0xa,
    0xa,  0xa,  0xa,  0xa,  0xa,  0xa,  0xa,  0xa,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,  0xb,
    0xb,  0xb,  0xb,  0xb,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,
    0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xc,  0xd,  0xd,  0xd,  0xd,
    0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,
    0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xd,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,
    0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,
    0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,
    0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xe,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,
    0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,
    0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,
    0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0xf,  0x0,  0x0, 0x10, 0x11,
   0x12, 0x12, 0x13, 0x13, 0x14, 0x14, 0x14, 0x14, 0x15, 0x15, 0x15, 0x15, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16,
   0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
   0x18, 0x18, 0x18, 0x18, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19,
   0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a,
   0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
   0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
   0x1b, 0x1b, 0x1b, 0x1b, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
   0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
   0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c,
   0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d,
   0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d,
   0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d,
   0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d
};

static const int d_extra_tab[30] = {
  0x0, 0x0, 0x0, 0x0, 0x1, 0x1, 0x2, 0x2, 0x3, 0x3, 0x4, 0x4, 0x5, 0x5,
  0x6, 0x6, 0x7, 0x7, 0x8, 0x8, 0x9, 0x9, 0xa, 0xa, 0xb, 0xb, 0xc, 0xc,
  0xd, 0xd
};

static const UBYTE srv_order[19] = {
   0x10, 0x11, 0x12, 0x0, 0x8, 0x7, 0x9, 0x6, 0xa, 0x5, 0xb, 0x4, 0xc, 0x3,
   0xd,  0x2,  0xe, 0x1, 0xf
};
