#/usr/bin/env bash

##
##                  INTEL CORPORATION PROPRIETARY INFORMATION
##     This software is supplied under the terms of a license agreement or
##     nondisclosure agreement with Intel Corporation and may not be copied
##     or disclosed except in accordance with the terms of that agreement.
##          Copyright(c) 2007-2010 Intel Corporation. All Rights Reserved.
##

PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/X11R6/bin:/root/bin:/usr/local/sbin:/usr/local/bin:/usr/sfw/bin:/usr/ccs/bin
export CC CXX CC_NAME COMPILER_HOME
export ICCLIBPATH=.

# ----------------------------------------------------------------------------------------------
# Setting info about sample
IPPVER="7.0"

#
SAMPLE_ROOT=`pwd`
SAMPLE_NAME=${SAMPLE_ROOT##*/}
SAMPLE_OS=`uname -s | tr [:upper:] [:lower:]`
SAMPLE_DATE=`date +%Y%m%d`
SAMPLE_SCRIPT=${0##*/}
SAMPLE_STATUS="-FAILED"

trap_list="0 1 3 6 9"

# ----------------------------------------------------------------------------------------------
# Functions
#
# Common samples header
# Usage		: type_header <msg>
#
type_header()
{
#O#
	sample_name=$1
	sample_dir=$2
	sample_script=$3
	sample_date=`date '+%a %b %d %r %Y'`
	echo "+----------------------------------------------------------------------------+"
	echo "|Test        : $sample_name build                $sample_date"
	echo "|Hostname    : "`hostname`
	echo "|Function    : no"
	echo "|Description : building $sample_name sample"
	echo "|Class       : Information"
	echo "|Source      : $sample_dir"
	echo "|Executable  : $sample_script"
	echo "+----------------------------------------------------------------------------+"
#C#
}
#
# Common samples footer
# Usage		: type_footer <msg>
#
type_footer()
{
#O#
	sample_name=$1
	sample_status=$2
	sample_date=`date '+%a %b %d %r %Y'`

	echo
	echo "+----------------------------------------------------------------------------+"
	echo "|$sample_status    $sample_name                   $sample_date"
	echo "+----------------------------------------------------------------------------+"

	if [ -f ${SAMPLE_BLK_FILE} ]; then rm -f ${SAMPLE_BLK_FILE}; fi
	trap - $trap_list

        ret=0
        if [ "$sample_status" != "-OK" ]; then ret=255; fi
        exit $ret
#C#
}
#
# print_error_msg
# Usage		: print_error_msg <msg>
#
print_error_msg()
{
#O#
	SAMPLE_STATUS="-FAILED"

	for((;;)); do
		[ -z "$1" ] && exit 255
		echo "ERROR: $1"
		shift
	done
#C#
}
#
# Execition checking
# Usage		: check_exec_mode <comp>
#
check_exec_mode()
{
#O#
	[ -z "$1" ] && { shift; print_error_msg "$@"; }

	[ -x "$1" ] && return
	shift
	print_error_msg "$@"
#C#
}

#
# Work checking
# Usage		: check_compiler <comp>
#
check_compiler()
{
#O#
    comp=$1
    cfile=/tmp/test-$$.c
    ofile=/tmp/test-$$.o
    rm -f $cfile $ofile
    
    echo 'int main(int argc, char** argv) { return 0; }' > $cfile
    ${comp} -c $cfile -o $ofile > /dev/null 2>&1 || print_error_msg "Unable to launch compiler properly or bad license"
    rm -f $cfile $ofile
#C#
}

#
#  detect_compiler
# Usage		: detect_compiler <comp_name> <mode>
#
detect_compiler()
{
#O#
	COMPILER_NAME="$1"
	mode="$2"

	_CC=icc
	_CXX=icpc

	case ${COMPILER_NAME} in
	    icc120  ) _VARS=compilervars ;;
	    *       ) _VARS=iccvars      ;;
	esac

	case "${SAMPLE_ARCH}" in
	    intel64 ) CCE=cce; FCE=fce ;;
	    ia32    ) CCE=cc;  FCE=fc  ;;
	esac
	
	                             
	COMPILER_HOME2=/Developer/opt/intel/composerxe-2011

	case ${COMPILER_NAME} in
            icc120      ) COMPILER_HOME=/opt/intel/composerxe-2011 ;;
	    icc111	) COMPILER_HOME=/opt/intel/Compiler/11.1/ ;;
	    icc110	) COMPILER_HOME=/opt/intel/Compiler/11.0/ ;;
	    icc101	) COMPILER_HOME=/opt/intel/${CCE}/10.1    ;;
	    icc10	) COMPILER_HOME=/opt/intel/${CCE}/10.0    ;;
	    icc91	) COMPILER_HOME=/opt/intel/${CCE}/9.1     ;;

	    ifc111	) COMPILER_HOME=/opt/intel/Compiler/11.1/ ;;
	    ifc110	) COMPILER_HOME=/opt/intel/Compiler/11.0/ ;;
	    ifc101	) COMPILER_HOME=/opt/intel/${FCE}/10.1    ;;
	    ifc10	) COMPILER_HOME=/opt/intel/${FCE}/10.0    ;;
	    ifc91	) COMPILER_HOME=/opt/intel/${FCE}/9.1     ;;

	    gfortran	) COMPILER_HOME=/usr ;;

	    gcc[34]*	) 
		COMPILER_HOME=/usr 
		case $SAMPLE_OS in
		    darwin	)
			[ "${COMPILER_NAME}" = "gcc3" ] && {
			    echo " Unsupported compiler version"
			    SAMPLE_STATUS="-UNSUPPORTED"
			    exit 
			}
		    ;;
		    sunos	) COMPILER_HOME=/usr/sfw ;;
		esac
	    ;;

	    custom	) ;;
	    *		)
		echo " Unsupported compiler version"
		SAMPLE_STATUS="-UNSUPPORTED"
		exit
	    ;;
	esac

	case ${COMPILER_NAME} in
	    ifc*	) _CC=ifort; _CXX=ifort; _VARS=ifortvars ;;
	esac

	case ${COMPILER_NAME} in
	    gcc[34]	)
		    GCC_VERSION=${COMPILER_NAME/gcc/}
		    [ -z ${GCC_HOME} ] && GCC_HOME=${COMPILER_HOME}
		    [ -z ${GCC_NAME} ] && GCC_NAME=gcc
		    [ -z ${GXX_NAME} ] && GXX_NAME=g++

		    COMPILER_HOME=${GCC_HOME}
		    GCC_HOME=${GCC_HOME}/bin
		    PATH=$GCC_HOME:$PATH

		    CC=${GCC_HOME}/${GCC_NAME}
		    CXX=${GCC_HOME}/${GXX_NAME}

		    [ -x ${CC} ] && {
			GCC_CUR=`${CC} -v 2>&1 | tail -1 | awk '{ print $3 }'`
			[ "${GCC_VERSION}" -eq "${GCC_CUR%%.*}" ] || CC=empty
		    } || CC=empty

		    [ -x ${CXX} ] && {
			GCC_CUR=`${CXX} -v 2>&1 | tail -1 | awk '{ print $3 }'`
			[ "${GCC_VERSION}" -eq "${GCC_CUR%%.*}" ] || CXX=empty
		    } || CXX=empty

		    [ "${CC}" = "empty" -a "${mode}" = "target" ] && {
			echo " *"
			echo " * You can use GCC_HOME, GCC_NAME and GXX_NAME"
			echo " * for detailing of gcc compiler specification."
			echo " *"
		    }
	    ;;

	    gfortran	)
		[ -z ${GCC_HOME} ] && GCC_HOME=${COMPILER_HOME}
		[ -z ${GCC_NAME} ] && GCC_NAME=gfortran
		[ -z ${GXX_NAME} ] && GXX_NAME=gfortran

		COMPILER_HOME=${GCC_HOME}
		GCC_HOME=${GCC_HOME}/bin
		PATH=$GCC_HOME:$PATH

		CC=${GCC_HOME}/${GCC_NAME}
		CXX=${GCC_HOME}/${GXX_NAME}

		[ "${CC}" = "empty" -a "${mode}" = "target" ] && {
		    echo " *"
		    echo " * You can use GCC_HOME, GCC_NAME and GXX_NAME"
		    echo " * for detailing of gcc compiler specification."
		    echo " *"
		}
	    ;;

	    icc* | ifc*	)
		COMPILER_HOME=`ls -d ${COMPILER_HOME}* 2>/dev/null | sort -r | head -1`
		CCFIND=`ls -d ${COMPILER_HOME}* | sort -r | head -1`/bin/$_VARS.sh
		
		[ -x "${CCFIND}" ] || {
		    COMPILER_HOME=`ls -d ${COMPILER_HOME2}* 2>/dev/null | sort -r | head -1`
		    CCFIND=`ls -d ${COMPILER_HOME}* | sort -r | head -1`/bin/$_VARS.sh
		}
		
		[ -x "${CCFIND}" ] || [ -r "${CCFIND}" ] && CCENV=${CCFIND}
		[ -z "${CCENV}" ] || {
		    . ${CCENV} ${SAMPLE_ARCH}
		    CC=`\which $_CC 2>/dev/null`
		    CXX=`\which $_CXX 2>/dev/null`
		}
            ;;

	    custom	)
		CC=${CC}
		CXX=${CXX}
		CC_NAME=${COMPILER_NAME}
		COMPILER_HOME=${COMPILER_HOME}
		CCENV=${CCENV}

		[ -z "${CCENV}" ] || . ${CCENV}
	    ;;
	esac

	if [ -x "${CC}" -a -x "${CXX}" ]; then CC_NAME=${COMPILER_NAME}; fi
#C#
}

#
# Checking OMP
# Usage		: detect_OMP <comp> <msg>
#
detect_OMP()
{
#O#
	case $CC_NAME in
	    gcc*	) ;;
	    icc*	) export OPENMP_SUPPORT=YES; return ;;
#	    *		) return ;;
	esac

	comp=$1
	check_exec_mode $comp

	export OPENMP_SUPPORT=NO

	cfile=/tmp/test-$$.c
	ofile=/tmp/test-$$.o
	rm -f $cfile $ofile

	echo '
	#include <omp.h>

	int main(int argc, char** argv) { int counter = 0;
	#pragma omp parallel num_threads(2) default(shared)
	{
	#pragma omp critical
		{ printf("%d", counter++); }
	} return 0; } ' > $cfile

	${comp} -fopenmp -c $cfile -o $ofile > /dev/null 2>&1 && OPENMP_SUPPORT=YES

	rm -f $cfile $ofile

	echo " OpenMP support for this compiler: ${OPENMP_SUPPORT}"
#C#
}

#
# Checking Qt
# Usage		: detect_Qt <comp> <ver> <msg>
#
detect_Qt()
{
#O#
	ver=$2
	mj=$3
	
	case $ver in
	    3 ) QT_include='#include <qglobal.h>' ;;
	    4 ) QT_include='#include <Qt/qglobal.h>' ;;
	esac

	[ -z "$QTDIR" -o ! -d "$QTDIR" ] && {
	    case $ver in
		3 )
		    for i in qt-3 qt3 ; do
			QTDIR=/usr/lib/$i
			[ "$SAMPLE_ARCH" ==  "intel64" ] && QTDIR=/usr/lib64/$i
			QTDIR=`ls -d ${QTDIR}* 2>/dev/null | sort -r | head -1`
			[ -z $QTDIR ] || break
		    done
		;;

		4 ) QTDIR=`ls -d /usr/local/Trolltech/Qt-4* 2>/dev/null | sort -r | head -1` ;;
		* ) print_error_msg "Requred Qt version (ver $ver) doesn't support yet.";;
	    esac
	}

	[ -d $QTDIR ] || print_error_msg "Requred Qt version (ver $ver) is not detected." "Please install"

	comp=$1
	check_exec_mode $comp
	check_compiler $comp

	cfile=/tmp/test-$$.c
	ofile=/tmp/test-$$.o
	rm -f $cfile $ofile

	echo "
	$QT_include

	#if ! defined (QT_VERSION_CHECK)
	#define QT_VERSION_CHECK(V,M,P) (((V)<<16) + ((M)<<8) + (P))
	#endif

	int main(void) {
		return
	#if (QT_VERSION >= QT_VERSION_CHECK($ver, 0, 0))
		0
	#else
		1
	#endif
		;
	}" > $cfile

	${comp} -I$QTDIR/include $cfile -o $ofile > /dev/null 2>&1 || print_error_msg "Qt: Qt (ver ${ver}.${mj}.x or later) is not detected in system. Please install"
	$ofile || print_error_msg print_error_msg "Qt: Qt (ver ${ver}.${mj}.x or later) is not detected in system. Please install"
	rm -f $cfile $ofile

	export QTDIR
	echo " Qt: Qt (ver $ver or later) is detected"
#C#
}
#
# ----------------------------------------------------------------------------------------------
# Setting up samples ENV
#

trap "type_footer $SAMPLE_NAME \$SAMPLE_STATUS" $trap_list
type_header $SAMPLE_NAME $SAMPLE_DIR $SAMPLE_SCRIPT $SAMPLE_DATE

# 
# Checking Linkage mode
# 
case "${LINKAGE}" in
	static | dynamic ) LINKAGE=${LINKAGE} ;;
	*		 ) LINKAGE=dynamic ;;
esac

case "${SAMPLE_OS}" in
	sunos		 ) LINKAGE=static ;;
esac

export LINKAGE

# 
# Compiler detection
#

case $SAMPLE_NAME in 
	f90* ) compiler_list="ifc111 ifc110 gfortran" ;;
	   * ) compiler_list="icc120 icc111 icc110 gcc4 gcc3" ;;
esac

[ -z "$1" ] && ARG=autodetect

case $ARG in
	auto* )
	for i in $compiler_list ; do
		GCC_HOME='' ; GCC_NAME='' ; GXX_NAME=''
		detect_compiler $i ${ARG}
		[ -z "${CC_NAME}" ] || break
	done ;;

	* ) detect_compiler $ARG target ;;
esac

check_exec_mode "${CC}"  " Compiler is installed non by default setting. Please reinstall."
check_exec_mode "${CXX}" " Compiler is installed non by default setting. Please reinstall."
[ -z "${CC_NAME}" ] && print_error_msg  "Compiler is installed non by default setting. Please reinstall."

echo " Found compiler ${CC_NAME} ( ${ARG} ) is ${COMPILER_HOME}"
echo " Compiler environment is loaded"

# 
# OMP mode detection
# 
detect_OMP $CC
 
# 
# Checkinh requred package
#

case $SAMPLE_OS in
	darwin	) SKIP_QT3=yes ;;
esac 

case $SAMPLE_NAME in
	jpegview ) [ -z "$SKIP_QT3" ] && detect_Qt $CC 3 0  ;;
	uic	 ) [ -z "$SKIP_QT4" ] && detect_Qt $CC 4 3  ;;
	*	 ) ;;
esac

# 
# checking IPP version

SAMPLE_IPPROOT_DEV=/Developer/opt/intel/composerxe-2011*
case $SAMPLE_OS in
        linux  | darwin ) SAMPLE_IPPROOT=/opt/intel/composerxe-2011* ;; 
	* ) SAMPLE_STATUS=-UNSUPPORTED
            print_error_msg  "This OS ($SAMPLE_OS) is not supported"
        ;;
esac

SAMPLE_IPPENV=ippvars.sh
[ -z "$IPPENV" ] && IPPENV=$SAMPLE_IPPENV
[ -f "$IPPENV" -o ! -f "$IPPENV" ] && IPPENV=$IPPROOT/bin/$SAMPLE_IPPENV
[ -z "$IPPENV" -o ! -f "$IPPENV" ] && IPPENV=`ls $SAMPLE_IPPROOT/ipp/bin/$SAMPLE_IPPENV 2>/dev/null | head -1`
[ -z "$IPPENV" -o ! -f "$IPPENV" ] && IPPENV=`ls $SAMPLE_IPPROOT_DEV/ipp/bin/$SAMPLE_IPPENV 2>/dev/null | head -1`

[ -z "$IPPENV" -a ! -x "$IPPENV" ] && print_error_msg " " \
	"***************************************************************************" \
	" Intel(R) IPP $IPPVER is not detected in your system." \
	" Please install Intel(R) IPP or set IPPROOT environment variable correctly " \
	"***************************************************************************"

echo " IPP environment found at $IPPENV"
. $IPPENV $SAMPLE_ARCH
echo " IPP environment is loaded"

#
# Some fix and trick
#
export LIBPTHREAD=-lpthread
OS_EXT=.dylib
LD_LIST=$DYLD_LIBRARY_PATH

case $SAMPLE_OS in
	linux 	) 
	    OS_EXT=.so
	    LD_LIST=$LD_LIBRARY_PATH
	;;
esac

#
# Search libiomp5.so lib
#
[ -z "$LD_LIST" ] || {
	save_IFS=$IFS
	IFS=: LD_ARR=($LD_LIST)
	IFS=$save_IFS
	for i in ${LD_ARR[@]}; do 
		[ -e $i/libiomp5$OS_EXT ] && LIBIOMP5=${LIBIOMP5:=-L$i}
	done

	[ -z "$LIBIOMP5" ] || LIBPTHREAD="$LIBIOMP5 -liomp5 $LIBPTHREAD"
}

# 
# Duplication execution is blocked
#

SAMPLE_BLK_FILE=${SAMPLE_DIR}/${ARCH}_${CC_NAME}.lck

if [ -f ${SAMPLE_BLK_FILE} ]; then
	SAMPLE_BLK_FILE=${SAMPLE_BLK_FILE}_empty
	type_footer $SAMPLE_NAME "-DUPLICATE"
	exit
fi

( hostname; date ) > ${SAMPLE_BLK_FILE}

echo " All necessary packages and requred compilers are checked"
echo " Environment is setup. Build process is started. Please wait build results ..."
echo

# vim: fmr=#O#,#C#
